@@ -1,30 +1,31 @@
-#!/usr/bin/env python3
-
 # std
+from __future__ import annotations
+
 import copy
+import pathlib
 import time
+from contextlib import closing
+from sqlite3 import Connection
+from typing import Any, Iterable, Sequence
 
 # 3rd
 import numpy as np
 import pandas as pd
-import pathlib
-from typing import Union, List, Dict, Optional, Iterable
+
+import ankipandas._columns as _columns
 
 # ours
-import ankipandas.paths
 import ankipandas.raw as raw
-import ankipandas.util.dataframe
-from ankipandas.util.dataframe import replace_df_inplace
-import ankipandas._columns as _columns
-from ankipandas.util.misc import invert_dict, flatten_list_list
-from ankipandas.util.log import log
 from ankipandas.util.checksum import field_checksum
+from ankipandas.util.dataframe import merge_dfs, replace_df_inplace
 from ankipandas.util.guid import guid as generate_guid
+from ankipandas.util.log import log
+from ankipandas.util.misc import flatten_list_list, invert_dict
 from ankipandas.util.types import (
-    is_list_list_like,
-    is_list_like,
-    is_list_dict_like,
     is_dict_list_like,
+    is_list_dict_like,
+    is_list_like,
+    is_list_list_like,
 )
 
 
@@ -42,13 +43,14 @@ class AnkiDataFrame(pd.DataFrame):
     ]
 
     def __init__(self, *args, **kwargs):
-        """ Initializes a blank :class:`AnkiDataFrame`.
+        """Initializes a blank :class:`AnkiDataFrame`.
 
         .. warning::
 
             It is recommended to directly initialize this class with the notes,
             cards or revs table, using one of the methods
-            :meth:`.notes`, :meth:`.cards` or :meth:`.revs` instead!
+            :meth:`.notes`, :meth:`.cards` or :meth:`.revs` of the
+            :class:`~ankipandas.collection.Collection` class instead!
 
         Args:
             *args: Internal use only. See arguments of
@@ -69,7 +71,7 @@ class AnkiDataFrame(pd.DataFrame):
         # gets set by _get_table
         #: Type of anki table: 'notes', 'cards' or 'revlog'. This corresponds to
         #: the meaning of the ID row.
-        self._anki_table = None  # type: str
+        self._anki_table: str = None
 
         #: Prefix for fields as columns. Default is ``nfld_``.
         self.fields_as_columns_prefix = "nfld_"
@@ -80,12 +82,12 @@ class AnkiDataFrame(pd.DataFrame):
 
         # gets set by _get_table
         # noinspection PyTypeChecker
-        #: Overal structure of the dataframe ``anki``, ``ours``, ``in_progress``
-        self._df_format = None  # type: str
+        #: Overall structure of the dataframe ``anki``, ``ours``, ``in_progress``
+        self._df_format: str = None
 
     @property
     def _constructor(self):
-        """ This needs to be overriden so that any DataFrame operations do not
+        """This needs to be overridden so that any DataFrame operations do not
         return a :class:`pandas.DataFrame` but a :class:`AnkiDataFrame`."""
 
         return AnkiDataFrame
@@ -101,7 +103,8 @@ class AnkiDataFrame(pd.DataFrame):
         if empty:
             df = raw.get_empty_table(table)
         else:
-            df = raw.get_table(col.db, table)
+            with closing(col.db) as db:
+                df = raw.get_table(db, table)
 
         replace_df_inplace(self, df)
         self.normalize(inplace=True)
@@ -157,7 +160,7 @@ class AnkiDataFrame(pd.DataFrame):
                             f1=self._fields_format, f2=other._fields_format
                         )
                     )
-        super(AnkiDataFrame, self).update(other, **kwargs)
+        super().update(other, **kwargs)
         # Fix https://github.com/pandas-dev/pandas/issues/4094
         for col, typ in _columns.dtype_casts2[self._anki_table].items():
             self[col] = self[col].astype(typ)
@@ -169,16 +172,15 @@ class AnkiDataFrame(pd.DataFrame):
         duplicates = self.index[self.index.duplicated()].tolist()
         if duplicates:
             log.critical(
-                "Duplicated indizes in table {} discovered, so something "
+                "Duplicated indizes in table %s discovered, so something "
                 "definitely went wrong. Please don't ignore this warning. "
-                "These indizes appear more "
-                "than once: {}".format(
-                    self._anki_table, ", ".join(map(str, duplicates))
-                )
+                "These indizes appear more than once: %s",
+                self._anki_table,
+                ", ".join(map(str, duplicates)),
             )
 
     def _invalid_table(self):
-        raise ValueError("Invalid table: {}.".format(self._anki_table))
+        raise ValueError(f"Invalid table: {self._anki_table}.")
 
     def _check_df_format(self):
         if self._df_format == "in_progress":
@@ -193,9 +195,7 @@ class AnkiDataFrame(pd.DataFrame):
         elif self._df_format == "ours":
             pass
         else:
-            raise ValueError(
-                "Unknown value of _df_format: {}".format(self._df_format)
-            )
+            raise ValueError(f"Unknown value of _df_format: {self._df_format}")
 
     def _check_our_format(self):
         self._check_df_format()
@@ -211,8 +211,11 @@ class AnkiDataFrame(pd.DataFrame):
     # ==========================================================================
 
     @property
-    def db(self):
-        """Opened Anki database (:class:`sqlite3.Connection`)"""
+    def db(self) -> Connection:
+        """Opened Anki database (:class:`sqlite3.Connection`). Make sure to
+        call `db.close()` after you're done. Better still, use
+        `contextlib.closing`.
+        """
         return self.col.db
 
     # IDs
@@ -220,8 +223,7 @@ class AnkiDataFrame(pd.DataFrame):
 
     @property
     def id(self):
-        """ Return note/card/review ID as :class:`pandas.Series` of integers.
-        """
+        """Return note/card/review ID as :class:`pandas.Series` of integers."""
         if self._anki_table == "notes":
             return self.nid
         elif self._anki_table == "cards":
@@ -233,7 +235,7 @@ class AnkiDataFrame(pd.DataFrame):
 
     @property
     def nid(self):
-        """ Note ID as :class:`pandas.Series` of integers. """
+        """Note ID as :class:`pandas.Series` of integers."""
         if self._anki_table == "notes":
             return self.index
         elif self._anki_table == "cards":
@@ -265,7 +267,7 @@ class AnkiDataFrame(pd.DataFrame):
 
     @property
     def cid(self):
-        """ Card ID as :class:`pandas.Series` of integers. """
+        """Card ID as :class:`pandas.Series` of integers."""
         if self._anki_table == "cards":
             return self.index
         if self._anki_table == "revs":
@@ -302,7 +304,7 @@ class AnkiDataFrame(pd.DataFrame):
 
     @property
     def rid(self):
-        """ Review ID as :class:`pandas.Series` of integers. """
+        """Review ID as :class:`pandas.Series` of integers."""
         if self._anki_table == "revs":
             return self.index
         else:
@@ -331,7 +333,7 @@ class AnkiDataFrame(pd.DataFrame):
 
     @property
     def mid(self):
-        """ Model ID as :class:`pandas.Series` of integers. """
+        """Model ID as :class:`pandas.Series` of integers."""
         if self._anki_table in ["notes"]:
             if "nmodel" not in self.columns:
                 raise ValueError(
@@ -360,7 +362,7 @@ class AnkiDataFrame(pd.DataFrame):
 
     @property
     def did(self):
-        """ Deck ID as :class:`pandas.Series` of integers. """
+        """Deck ID as :class:`pandas.Series` of integers."""
         if self._anki_table == "cards":
             if "cdeck" not in self.columns:
                 raise ValueError(
@@ -390,7 +392,7 @@ class AnkiDataFrame(pd.DataFrame):
 
     @property
     def odid(self):
-        """ Original deck ID for cards in filtered deck as
+        """Original deck ID for cards in filtered deck as
         :class:`pandas.Series` of integers.
         """
         if self._anki_table == "cards":
@@ -432,7 +434,7 @@ class AnkiDataFrame(pd.DataFrame):
         prepend="n",
         prepend_clash_only=True,
     ):
-        """ Merge note table into existing dataframe.
+        """Merge note table into existing dataframe.
 
         Args:
             inplace: If False, return new dataframe, else update old one
@@ -453,7 +455,7 @@ class AnkiDataFrame(pd.DataFrame):
             )
         elif self._anki_table == "revs":
             self["nid"] = self.nid
-        ret = ankipandas.util.dataframe.merge_dfs(
+        ret = merge_dfs(
             df=self,
             df_add=self.col.notes,
             id_df="nid",
@@ -500,7 +502,7 @@ class AnkiDataFrame(pd.DataFrame):
                 "notes table."
             )
         self._check_our_format()
-        ret = ankipandas.util.dataframe.merge_dfs(
+        ret = merge_dfs(
             df=self,
             df_add=self.col.cards,
             id_df="cid",
@@ -551,9 +553,7 @@ class AnkiDataFrame(pd.DataFrame):
         elif self._fields_format == "list":
             pass
         else:
-            raise ValueError(
-                "Unknown _fields_format: {}".format(self._fields_format)
-            )
+            raise ValueError(f"Unknown _fields_format: {self._fields_format}")
 
         if "nflds" not in self.columns:
             raise ValueError("Could not find fields column 'nflds'.")
@@ -613,9 +613,7 @@ class AnkiDataFrame(pd.DataFrame):
         elif self._fields_format == "columns":
             pass
         else:
-            raise ValueError(
-                "Unknown _fields_format: {}".format(self._fields_format)
-            )
+            raise ValueError(f"Unknown _fields_format: {self._fields_format}")
 
         self._fields_format = "in_progress"
         mids = self.mid.unique()
@@ -643,8 +641,8 @@ class AnkiDataFrame(pd.DataFrame):
                 "the notes into your table?"
             )
 
-    def list_tags(self) -> List[str]:
-        """ Return sorted list of all tags in the current table. """
+    def list_tags(self) -> list[str]:
+        """Return sorted list of all tags in the current table."""
         if "ntags" not in self.columns:
             raise ValueError(
                 "Tags column 'ntags' not present. Either use the notes table"
@@ -652,37 +650,33 @@ class AnkiDataFrame(pd.DataFrame):
             )
         else:
             return sorted(
-                list(
-                    set(
-                        [item for lst in self["ntags"].tolist() for item in lst]
-                    )
-                )
+                {item for lst in self["ntags"].tolist() for item in lst}
             )
 
-    def list_decks(self) -> List[str]:
-        """ Return sorted list of deck names in the current table. """
+    def list_decks(self) -> list[str]:
+        """Return sorted list of deck names in the current table."""
         if "cdeck" not in self.columns:
             raise ValueError(
                 "Deck column 'cdeck' not present. Either use the cards table "
                 "or merge it into your table."
             )
         else:
-            decks = sorted(list(self["cdeck"].unique()))
+            decks = sorted(self["cdeck"].unique())
             if "" in decks:
                 decks.remove("")
             return decks
 
     def list_models(self):
-        """ Return sorted list of model names in the current table. """
+        """Return sorted list of model names in the current table."""
         if "nmodel" not in self.columns:
             raise ValueError(
                 "Model column 'nmodel' not present. Either use the notes table"
                 " or merge it into your table."
             )
-        return sorted(list(self["nmodel"].unique()))
+        return sorted(self["nmodel"].unique())
 
-    def has_tag(self, tags: Optional[Union[Iterable[str], str]] = None):
-        """ Checks whether row has a certain tag ('ntags' column).
+    def has_tag(self, tags: Iterable[str] | str | None = None):
+        """Checks whether row has a certain tag ('ntags' column).
 
         Args:
             tags: String or list thereof. In the latter case, True is returned
@@ -722,8 +716,8 @@ class AnkiDataFrame(pd.DataFrame):
         else:
             return self["ntags"].apply(bool)
 
-    def has_tags(self, tags: Optional[Union[Iterable[str], str]] = None):
-        """ Checks whether row contains at least the supplied tags.
+    def has_tags(self, tags: Iterable[str] | str | None = None):
+        """Checks whether row contains at least the supplied tags.
 
         Args:
             tags: String or list thereof.
@@ -751,8 +745,8 @@ class AnkiDataFrame(pd.DataFrame):
         _has_tags = set(tags).issubset
         return self["ntags"].apply(_has_tags)
 
-    def add_tag(self, tags: Union[Iterable[str], str], inplace=False):
-        """ Adds tag ('ntags' column).
+    def add_tag(self, tags: Sequence[str] | str, inplace=False):
+        """Adds tag ('ntags' column).
 
         Args:
             tags: String or list thereof.
@@ -775,12 +769,12 @@ class AnkiDataFrame(pd.DataFrame):
             return
 
         def _add_tags(other):
-            return other + sorted(list(set(tags) - set(other)))
+            return other + sorted(set(tags) - set(other))
 
         self["ntags"] = self["ntags"].apply(_add_tags)
 
-    def remove_tag(self, tags: Union[Iterable[str], str, None], inplace=False):
-        """ Removes tag ('ntags' column).
+    def remove_tag(self, tags: Iterable[str] | str | None, inplace=False):
+        """Removes tag ('ntags' column).
 
         Args:
             tags: String or list thereof. If None, all tags are removed.
@@ -813,9 +807,9 @@ class AnkiDataFrame(pd.DataFrame):
     # ==========================================================================
 
     def was_modified(
-        self, other: Optional[pd.DataFrame] = None, na=True, _force=False
+        self, other: pd.DataFrame | None = None, na=True, _force=False
     ):
-        """ Compare with original table, show which rows have changed.
+        """Compare with original table; show which rows have changed.
         Will only compare columns existing in both dataframes.
 
         Args:
@@ -832,23 +826,24 @@ class AnkiDataFrame(pd.DataFrame):
             self._check_our_format()
 
         if other is None:
-            other = self.col._get_original_item(
+            _other: AnkiDataFrame = self.col._get_original_item(
                 self._anki_table
-            )  # type: AnkiDataFrame
+            )
+        else:
+            _other = other
+        del other  # avoid confusion
 
         self_sf = self
         if self._fields_format == "columns":
             self_sf = self.fields_as_list(inplace=False, force=_force)
 
-        cols = sorted(
-            list(set(self_sf.columns).intersection(set(other.columns)))
-        )
+        cols = sorted(set(self_sf.columns) & set(_other.columns))
 
-        other_nids = set(other.index)
-        inters = set(self_sf.index).intersection(other_nids)
+        other_nids = set(_other.index)
+        inters = set(self_sf.index) & other_nids
         result = pd.Series(na, index=self_sf.index)
         new_bools = np.any(
-            other.loc[other.index.isin(inters), cols].values
+            _other.loc[_other.index.isin(inters), cols].values
             != self_sf.loc[self_sf.index.isin(inters), cols].values,
             axis=1,
         )
@@ -858,9 +853,9 @@ class AnkiDataFrame(pd.DataFrame):
         return result
 
     def modified_columns(
-        self, other: Optional[pd.DataFrame] = None, _force=False, only=True
+        self, other: pd.DataFrame | None = None, _force=False, only=True
     ):
-        """ Compare with original table, show which columns in which rows
+        """Compare with original table, show which columns in which rows
         were modified.
 
         Args:
@@ -878,20 +873,22 @@ class AnkiDataFrame(pd.DataFrame):
             other = self.init_with_table(col=self.col, table=self._anki_table)
         cols = [c for c in self.columns if c in other.columns]
         other_nids = set(other.index)
-        inters = set(self.index).intersection(other_nids)
+        inters = set(self.index) & other_nids
         if only:
-            inters = inters.intersection(
+            inters &= set(
                 self[self.was_modified(other=other, _force=_force)].index
             )
-        inters = sorted(list(inters))
+        inters_st = sorted(inters)
+        del inters
         return pd.DataFrame(
-            self.loc[inters, cols].values != other.loc[inters, cols].values,
-            index=self.loc[inters].index,
+            self.loc[inters_st, cols].values
+            != other.loc[inters_st, cols].values,
+            index=self.loc[inters_st].index,
             columns=cols,
         )
 
-    def was_added(self, other: Optional[pd.DataFrame] = None, _force=False):
-        """ Compare with original table, show which rows were added.
+    def was_added(self, other: pd.DataFrame | None = None, _force=False):
+        """Compare with original table, show which rows were added.
 
         Args:
             other: Compare with this :class:`pandas.DataFrame`.
@@ -915,9 +912,9 @@ class AnkiDataFrame(pd.DataFrame):
         return self.index.isin(new_indices)
 
     def was_deleted(
-        self, other: Optional[pd.DataFrame] = None, _force=False
-    ) -> List:
-        """ Compare with original table, return deleted indizes.
+        self, other: pd.DataFrame | None = None, _force=False
+    ) -> list:
+        """Compare with original table, return deleted indizes.
 
         Args:
             other: Compare with this :class:`pandas.DataFrame`.
@@ -937,20 +934,20 @@ class AnkiDataFrame(pd.DataFrame):
             other_ids = set(self.col._get_original_item(self._anki_table).id)
 
         deleted_indices = other_ids - set(self.index)
-        return sorted(list(deleted_indices))
+        return sorted(deleted_indices)
 
     # Update modification stamps and similar
     # ==========================================================================
 
     def _set_usn(self):
-        """ Update usn (update sequence number) for all changed rows. """
+        """Update usn (update sequence number) for all changed rows."""
         self.loc[
             self.was_modified(na=True, _force=True),
             _columns.columns_anki2ours[self._anki_table]["usn"],
         ] = -1
 
     def _set_mod(self):
-        """ Update modification timestamps for all changed rows. """
+        """Update modification timestamps for all changed rows."""
         if self._anki_table in ["cards", "notes"]:
             self.loc[
                 self.was_modified(na=True, _force=True),
@@ -959,7 +956,7 @@ class AnkiDataFrame(pd.DataFrame):
 
     # todo: test
     def _set_guid(self):
-        """ Update globally unique id """
+        """Update globally unique id"""
         if self._anki_table == "notes":
             self.loc[~self["nguid"].apply(bool)].apply(generate_guid)
 
@@ -967,7 +964,7 @@ class AnkiDataFrame(pd.DataFrame):
     # ==========================================================================
 
     def normalize(self, inplace=False, force=False):
-        """ Bring a :class:`AnkiDataFrame` from the ``raw`` format (i.e. the
+        """Bring a :class:`AnkiDataFrame` from the ``raw`` format (i.e. the
         exact format that Anki uses in its internal representation) to our
         convenient format.
 
@@ -1030,9 +1027,8 @@ class AnkiDataFrame(pd.DataFrame):
         if duplicate_ids:
             log.critical(
                 "The following IDs occur "
-                "more than once: {}. Please do not use this dataframe.".format(
-                    ", ".join(map(str, duplicate_ids))
-                )
+                "more than once: %s. Please do not use this dataframe.",
+                ", ".join(map(str, duplicate_ids)),
             )
         self.set_index(id_field, inplace=True)
 
@@ -1070,7 +1066,7 @@ class AnkiDataFrame(pd.DataFrame):
         self._df_format = "ours"
 
     def raw(self, inplace=False, force=False):
-        """ Bring a :class:`AnkiDataFrame` into the ``raw`` format (i.e. the
+        """Bring a :class:`AnkiDataFrame` into the ``raw`` format (i.e. the
         exact format that Anki uses in its internal representation) .
 
         Args:
@@ -1210,8 +1206,8 @@ class AnkiDataFrame(pd.DataFrame):
     # Write
     # ==========================================================================
 
-    def summarize_changes(self, output="print") -> Optional[dict]:
-        """ Summarize changes that were made with respect to the table
+    def summarize_changes(self, output="print") -> dict | None:
+        """Summarize changes that were made with respect to the table
         as loaded from the database.
 
         Args:
@@ -1236,16 +1232,17 @@ class AnkiDataFrame(pd.DataFrame):
             print("Modified rows: {}".format(as_dict["n_modified"]))
             print("Added rows: {}".format(as_dict["n_added"]))
             print("Deleted rows: {}".format(as_dict["n_deleted"]))
+            return None  # make explicit for mypy
         elif output == "dict":
             return as_dict
         else:
-            raise ValueError("Invalid output setting: {}".format(output))
+            raise ValueError(f"Invalid output setting: {output}")
 
     # Append
     # ==========================================================================
 
     def _get_id(self, others=()) -> int:
-        """ Generate ID from timestamp and increment if it is already in use.
+        """Generate ID from timestamp and increment if it is already in use.
 
         .. warning::
 
@@ -1262,17 +1259,17 @@ class AnkiDataFrame(pd.DataFrame):
         self,
         nid: int,
         cdeck: str,
-        cord: Optional[Union[int, List[int]]] = None,
-        cmod: Optional[int] = None,
-        cusn: Optional[int] = None,
-        cqueue: Optional[str] = None,
-        ctype: Optional[str] = None,
-        civl: Optional[int] = None,
-        cfactor: Optional[int] = None,
-        creps: Optional[int] = None,
-        clapses: Optional[int] = None,
-        cleft: Optional[int] = None,
-        cdue: Optional[int] = None,
+        cord: int | list[int] | None = None,
+        cmod: int | None = None,
+        cusn: int | None = None,
+        cqueue: str | None = None,
+        ctype: str | None = None,
+        civl: int | None = None,
+        cfactor: int | None = None,
+        creps: int | None = None,
+        clapses: int | None = None,
+        cleft: int | None = None,
+        cdue: int | None = None,
         inplace=False,
     ):
         """
@@ -1317,21 +1314,22 @@ class AnkiDataFrame(pd.DataFrame):
     # todo: change order of arguments?
     # fixme: cord will be replaced
     # todo: duplicate cards (same note, same cord)?
+    # fixme: This is an absolute mess with the signature and mypy...
     def add_cards(
         self,
-        nid: List[int],
-        cdeck: Union[str, List[str]],
-        cord: Optional[Union[int, List[int]]] = None,
-        cmod: Optional[Union[int, List[int]]] = None,
-        cusn: Optional[Union[int, List[int]]] = None,
-        cqueue: Optional[Union[str, List[str]]] = None,
-        ctype: Optional[Union[str, List[str]]] = None,
-        civl: Optional[Union[int, List[int]]] = None,
-        cfactor: Optional[Union[int, List[int]]] = None,
-        creps: Optional[Union[int, List[int]]] = None,
-        clapses: Optional[Union[int, List[int]]] = None,
-        cleft: Optional[Union[int, List[int]]] = None,
-        cdue: Optional[Union[int, List[int]]] = None,
+        nid: list[int],
+        cdeck: str | list[str],
+        cord: int | list[int] | None = None,
+        cmod: int | list[int] | None = None,
+        cusn: int | list[int] | None = None,
+        cqueue: str | list[str] | None = None,
+        ctype: str | list[str] | None = None,
+        civl: int | list[int] | None = None,
+        cfactor: int | list[int] | None = None,
+        creps: int | list[int] | None = None,
+        clapses: int | list[int] | None = None,
+        cleft: int | list[int] | None = None,
+        cdue: int | list[int] | None = None,
         inplace=False,
     ):
         """
@@ -1396,7 +1394,7 @@ class AnkiDataFrame(pd.DataFrame):
         # --- Ord ---
 
         nid2mid = raw.get_nid2mid(self.db)
-        missing_nids = sorted(list(set(nid) - set(nid2mid.keys())))
+        missing_nids = sorted(set(nid) - set(nid2mid))
         if missing_nids:
             raise ValueError(
                 "The following note IDs (nid) can't be found in the notes "
@@ -1406,14 +1404,14 @@ class AnkiDataFrame(pd.DataFrame):
                 )
             )
 
-        mids = list(set(map(lambda x: nid2mid[x], nid)))
+        mids = {nid2mid[x] for x in nid}
         if len(mids) >= 2:
             raise ValueError(
                 "It is only supported to add cards for notes of the same model"
                 ", but you're trying to add cards for notes of "
                 "models: {}".format(", ".join(map(str, mids)))
             )
-        mid = mids[0]
+        mid = mids.pop()
 
         # fixme: should use function from ankipandas.raw
         available_ords = raw.get_mid2templateords(self.db)[mid]
@@ -1425,9 +1423,9 @@ class AnkiDataFrame(pd.DataFrame):
             pass
         else:
             raise ValueError(
-                "Unknown type for cord specifiation: {}".format(type(cord))
+                f"Unknown type for cord specifiation: {type(cord)}"
             )
-        not_available = sorted(list(set(cord) - set(available_ords)))
+        not_available = sorted(set(cord) - set(available_ords))
         if not_available:
             raise ValueError(
                 "The following templates are not available for notes of "
@@ -1443,12 +1441,12 @@ class AnkiDataFrame(pd.DataFrame):
                 raise ValueError(
                     "Number of decks doesn't match number of "
                     "notes for which cards should be added: {} "
-                    "instead of {}.".format(len(cmod), len(nid))
+                    "instead of {}.".format(len(cdeck), len(nid))
                 )
         else:
-            raise ValueError("Unknown format for cdeck: {}".format(type(cdeck)))
+            raise ValueError(f"Unknown format for cdeck: {type(cdeck)}")
         unknown_decks = sorted(
-            list(set(cdeck) - set(raw.get_did2deck(self.db).values()))
+            set(cdeck) - set(raw.get_did2deck(self.db).values())
         )
         if unknown_decks:
             raise ValueError(
@@ -1459,7 +1457,7 @@ class AnkiDataFrame(pd.DataFrame):
 
         # --- Rest ---
 
-        def _handle_input(inpt, name, default, typ, options=None):
+        def _handle_input(inpt, name, default, typ, options=None) -> list[Any]:
             if inpt is None:
                 inpt = [default] * len(nid)
             elif is_list_like(inpt):
@@ -1467,7 +1465,7 @@ class AnkiDataFrame(pd.DataFrame):
                     raise ValueError(
                         "Number of {} doesn't match number of "
                         "notes for which cards should be added: {} "
-                        "instead of {}.".format(name, len(cmod), len(nid))
+                        "instead of {}.".format(name, len(inpt), len(nid))
                     )
             elif isinstance(inpt, typ):
                 inpt = [inpt] * len(nid)
@@ -1478,7 +1476,7 @@ class AnkiDataFrame(pd.DataFrame):
                     )
                 )
             if options is not None:
-                invalid = sorted(list(set(inpt) - set(options)))
+                invalid = sorted(set(inpt) - set(options))
                 if invalid:
                     raise ValueError(
                         "The following values are no valid "
@@ -1530,17 +1528,17 @@ class AnkiDataFrame(pd.DataFrame):
                     )
                 )
         elif is_list_like(cdue):
-            if len(cdue) != len(nid):
+            if len(cdue) != len(nid):  # type: ignore
                 raise ValueError(
                     "Number of cdue doesn't match number of "
                     "notes for which cards should be added: {} "
-                    "instead of {}.".format(len(cmod), len(nid))
+                    "instead of {}.".format(len(cdue), len(nid))  # type: ignore
                 )
         elif isinstance(cdue, int):
             cdue = [cdue] * len(nid)
         else:
             raise ValueError(
-                "Invalid type of cdue specification: {}".format(type(cdue))
+                f"Invalid type of cdue specification: {type(cdue)}"
             )
 
         # Now we need to decide on contents for EVERY column in the DF
@@ -1583,14 +1581,14 @@ class AnkiDataFrame(pd.DataFrame):
             replace_df_inplace(self, self.append(add))
             return all_cids
 
-    def _get_ids(self, n=1) -> List[int]:
-        """ Generate ID from timestamp and increment if it is already in use.
+    def _get_ids(self, n=1) -> list[int]:
+        """Generate ID from timestamp and increment if it is already in use.
 
         Args:
             n: Number of IDs to generate
         """
-        indices = []
-        for i in range(n):
+        indices: list[int] = []
+        for _ in range(n):
             indices.append(self._get_id(others=indices))
         return indices
 
@@ -1598,15 +1596,15 @@ class AnkiDataFrame(pd.DataFrame):
     def add_notes(
         self,
         nmodel: str,
-        nflds: Union[List[List[str]], Dict[str, List[str]]],
-        ntags: List[List[str]] = None,
+        nflds: (list[list[str]] | dict[str, list[str]] | list[dict[str, str]]),
+        ntags: list[list[str]] | None = None,
         nid=None,
         nguid=None,
         nmod=None,
         nusn=None,
         inplace=False,
     ):
-        """ Add multiple new notes corresponding to one model.
+        """Add multiple new notes corresponding to one model.
 
         Args:
             nmodel: Name of the model (must exist already, check
@@ -1648,10 +1646,8 @@ class AnkiDataFrame(pd.DataFrame):
         # --- Model ---
 
         model2mid = raw.get_model2mid(self.db)
-        if nmodel not in model2mid.keys():
-            raise ValueError(
-                "No model of with name '{}' exists.".format(nmodel)
-            )
+        if nmodel not in model2mid:
+            raise ValueError(f"No model of with name '{nmodel}' exists.")
         field_keys = raw.get_mid2fields(self.db)[model2mid[nmodel]]
 
         # --- Fields ---
@@ -1659,19 +1655,18 @@ class AnkiDataFrame(pd.DataFrame):
         if is_list_dict_like(nflds):
             n_notes = len(nflds)
             specified_fields = set(
-                flatten_list_list(list(map(lambda d: list(d.keys()), nflds)))
-            )
-            unknown_fields = sorted(list(specified_fields - set(field_keys)))
+                flatten_list_list(list(map(lambda x: list(x), nflds)))
+            )  # mypy doesn't want to use list as a function here
+            unknown_fields = sorted(specified_fields - set(field_keys))
             if unknown_fields:
                 raise ValueError(
                     "Unknown fields: {}".format(", ".join(unknown_fields))
                 )
             field_key2field = {
-                key: list(map(lambda d: d.get(key), nflds))
-                for key in field_keys
+                key: [d.get(key) for d in nflds] for key in field_keys  # type: ignore
             }
         elif is_list_list_like(nflds):
-            n_fields = list(set(map(len, nflds)))
+            n_fields = list({len(x) for x in nflds})
             n_notes = len(nflds)
             if not (len(n_fields) == 1 and n_fields[0] == len(field_keys)):
                 raise ValueError(
@@ -1682,20 +1677,20 @@ class AnkiDataFrame(pd.DataFrame):
                     )
                 )
             field_key2field = {
-                field_keys[i]: list(map(lambda x: x[i], nflds))
-                for i in range(len(field_keys))
+                field_key: [x[i] for x in nflds]  # type: ignore
+                for i, field_key in enumerate(field_keys)
             }
         elif is_dict_list_like(nflds):
-            lengths = list(set(map(len, nflds.values())))
+            lengths = {len(x) for x in nflds.values()}  # type: ignore
             if len(lengths) >= 2:
                 raise ValueError(
                     "Inconsistent number of "
                     "fields: {}".format(", ".join(map(str, lengths)))
                 )
-            elif len(lengths) == 0:
+            elif not lengths:
                 raise ValueError("Are you trying to add zero notes?")
-            n_notes = lengths[0]
-            field_key2field = copy.deepcopy(nflds)
+            n_notes = lengths.pop()
+            field_key2field = copy.deepcopy(nflds)  # type: ignore
             for key in field_keys:
                 if key not in field_key2field:
                     field_key2field[key] = [""] * n_notes
@@ -1724,7 +1719,7 @@ class AnkiDataFrame(pd.DataFrame):
         else:
             nid = self._get_ids(n=n_notes)
 
-        already_present = sorted(list(set(nid).intersection(set(self.index))))
+        already_present = sorted(set(nid) & set(self.index))
         if already_present:
             raise ValueError(
                 "The following note IDs (nid) are "
@@ -1758,9 +1753,7 @@ class AnkiDataFrame(pd.DataFrame):
         else:
             nguid = [generate_guid() for _ in range(n_notes)]
 
-        existing_guids = sorted(
-            list(set(nguid).intersection(self["nguid"].unique()))
-        )
+        existing_guids = sorted(set(nguid) & set(self["nguid"].unique()))
         if existing_guids:
             raise ValueError(
                 "The following globally unique IDs (guid) are already"
@@ -1768,7 +1761,7 @@ class AnkiDataFrame(pd.DataFrame):
             )
 
         # todo: make efficient
-        duplicate_guids = sorted(set([g for g in nguid if nguid.count(g) >= 2]))
+        duplicate_guids = sorted({g for g in nguid if nguid.count(g) >= 2})
         if duplicate_guids:
             raise ValueError(
                 "The following gloally unique IDs (guid) are not unique: ",
@@ -1845,7 +1838,7 @@ class AnkiDataFrame(pd.DataFrame):
     def add_note(
         self,
         nmodel: str,
-        nflds: Union[List[str], Dict[str, str]],
+        nflds: list[str] | dict[str, str],
         ntags=None,
         nid=None,
         nguid=None,
@@ -1853,7 +1846,7 @@ class AnkiDataFrame(pd.DataFrame):
         nusn=-1,
         inplace=False,
     ):
-        """ Add new note.
+        """Add new note.
 
         .. note::
 
@@ -1887,14 +1880,17 @@ class AnkiDataFrame(pd.DataFrame):
             new note ID (``int``)
 
         """
+        _nflds: list[list[str]] | dict[str, list[str]] = []
         if is_list_like(nflds):
-            nflds = [nflds]
+            _nflds = [nflds]  # type: ignore
         elif isinstance(nflds, dict):
-            nflds = [nflds]
+            _nflds = {key: [value] for key, value in nflds.items()}
         else:
             raise ValueError(
-                "Unknown type for fields specification: {}".format(type(nflds))
+                f"Unknown type for fields specification: {type(nflds)}"
             )
+        del nflds
+
         if ntags is not None:
             ntags = [ntags]
         if nid is not None:
@@ -1908,7 +1904,7 @@ class AnkiDataFrame(pd.DataFrame):
 
         ret = self.add_notes(
             nmodel=nmodel,
-            nflds=nflds,
+            nflds=_nflds,
             ntags=ntags,
             nid=nid,
             nguid=nguid,
@@ -1927,7 +1923,7 @@ class AnkiDataFrame(pd.DataFrame):
     # ==========================================================================
 
     # todo: test?
-    def help_col(self, column, ret=False) -> Union[str, None]:
+    def help_col(self, column, ret=False) -> str | None:
         """
         Show description/help about a column. To get information about all
         columns, use the :meth:`.help_cols` method instead.
@@ -1938,16 +1934,14 @@ class AnkiDataFrame(pd.DataFrame):
         """
         df = self.help_cols(column)
         if len(df) == 0:
-            raise ValueError(
-                "Could not find help for your search request.".format(column)
-            )
+            raise ValueError("Could not find help for your search request.")
         if len(df) == 2:
             # fix for nid and cid column:
             df = self.help_cols(column, table=self._anki_table)
         if len(df) != 1:
             raise ValueError("Could not find help due to bug.")
         data = df.loc[column].to_dict()
-        h = "Help for column '{}'\n".format(column)
+        h = f"Help for column '{column}'\n"
         h += "-" * (len(h) - 1) + "\n"
         if data["Native"]:
             h += "Name in raw Anki database: " + data["AnkiColumn"] + "\n"
@@ -1958,6 +1952,7 @@ class AnkiDataFrame(pd.DataFrame):
             return h
         else:
             print(h)
+            return None  # Make explicit for mypy
 
     def help_cols(
         self, column="auto", table="all", ankicolumn="all"
@@ -2005,8 +2000,8 @@ class AnkiDataFrame(pd.DataFrame):
         return df
 
     @staticmethod
-    def help(ret=False) -> Union[str, None]:
-        """ Display short help text.
+    def help(ret=False) -> str | None:
+        """Display short help text.
 
         Args:
             ret: Return as string instead of printing it.
@@ -2030,3 +2025,4 @@ class AnkiDataFrame(pd.DataFrame):
             return h
         else:
             print(h)
+            return None  # explicit for mypy
