@@ -1,10 +1,11 @@
-#!/usr/bin/env python3
-
 # std
-from pathlib import Path, PurePath
+from __future__ import annotations
+
 import sqlite3
-from typing import Optional, Dict, Union, Any
 import time
+from contextlib import closing
+from pathlib import Path, PurePath
+from typing import Any
 
 # ours
 import ankipandas.paths
@@ -13,9 +14,9 @@ from ankipandas.ankidf import AnkiDataFrame
 from ankipandas.util.log import log
 
 
-class Collection(object):
+class Collection:
     def __init__(self, path=None, user=None):
-        """ Initialize :class:`~ankipandas.collection.Collection` object.
+        """Initialize :class:`~ankipandas.collection.Collection` object.
 
         Args:
             path: (Search) path to database. See
@@ -48,45 +49,34 @@ class Collection(object):
         path = ankipandas.paths.db_path_input(path, user=user)
 
         #: Path to currently loaded database
-        self._path = path  # type: Path
-
-        log.info("Loaded db from {}".format(self.path))
-
-        #: Opened Anki database (:class:`sqlite3.Connection`)
-        self._db = raw.load_db(self.path)  # type: sqlite3.Connection
+        self._path: Path = path
 
         #: Should be accessed with _get_item!
-        self.__items = {
+        self.__items: dict[str, AnkiDataFrame | None] = {
             "notes": None,
             "cards": None,
             "revs": None,
-        }  # type: Dict[str, Optional[AnkiDataFrame]]
+        }
 
         #: Should be accessed with _get_original_item!
-        self.__original_items = {
+        self.__original_items: dict[str, AnkiDataFrame | None] = {
             "notes": None,
             "cards": None,
             "revs": None,
-        }  # type: Dict[str, Optional[AnkiDataFrame]]
+        }
 
     @property
     def path(self) -> Path:
-        """ Path to currently loaded database """
+        """Path to currently loaded database"""
         return self._path
 
     @property
     def db(self) -> sqlite3.Connection:
-        """ Opened Anki database """
-        return self._db
-
-    def __del__(self):
-        log.debug(
-            "Closing db {db} which was loaded from {path}.".format(
-                db=self.db, path=self.path
-            )
-        )
-        raw.close_db(self.db)
-        log.debug("Closing successful")
+        """Opened Anki database. Make sure to call `db.close()` after you're
+        done. Better still, use `contextlib.closing`.
+        """
+        log.debug(f"Opening Db from {self._path}")
+        return raw.load_db(self._path)
 
     def _get_original_item(self, item):
         r = self.__original_items[item]
@@ -105,7 +95,7 @@ class Collection(object):
 
     @property
     def notes(self) -> AnkiDataFrame:
-        """ Notes as :class:`ankipandas.ankidf.AnkiDataFrame`. """
+        """Notes as :class:`ankipandas.ankidf.AnkiDataFrame`."""
         return self._get_item("notes")
 
     @notes.setter
@@ -114,7 +104,7 @@ class Collection(object):
 
     @property
     def cards(self) -> AnkiDataFrame:
-        """ Cards as :class:`ankipandas.ankidf.AnkiDataFrame`. """
+        """Cards as :class:`ankipandas.ankidf.AnkiDataFrame`."""
         return self._get_item("cards")
 
     @cards.setter
@@ -123,7 +113,7 @@ class Collection(object):
 
     @property
     def revs(self) -> AnkiDataFrame:
-        """ Reviews as :class:`ankipandas.ankidf.AnkiDataFrame`. """
+        """Reviews as :class:`ankipandas.ankidf.AnkiDataFrame`."""
         return self._get_item("revs")
 
     @revs.setter
@@ -131,22 +121,22 @@ class Collection(object):
         self.__items["revs"] = value
 
     def empty_notes(self):
-        """ Similar :class:`ankipandas.ankidf.AnkiDataFrame`
-        to :attr:`notes`, but without any rows. """
+        """Similar :class:`ankipandas.ankidf.AnkiDataFrame`
+        to :attr:`notes`, but without any rows."""
         return AnkiDataFrame.init_with_table(self, "notes", empty=True)
 
     def empty_cards(self):
-        """ Similar :class:`ankipandas.ankidf.AnkiDataFrame`
-        to :attr:`cards`, but without any rows. """
+        """Similar :class:`ankipandas.ankidf.AnkiDataFrame`
+        to :attr:`cards`, but without any rows."""
         return AnkiDataFrame.init_with_table(self, "cards", empty=True)
 
     def empty_revs(self):
-        """ Similar :class:`ankipandas.ankidf.AnkiDataFrame`
-        to :attr:`revs`, but without any rows. """
+        """Similar :class:`ankipandas.ankidf.AnkiDataFrame`
+        to :attr:`revs`, but without any rows."""
         return AnkiDataFrame.init_with_table(self, "revs", empty=True)
 
-    def summarize_changes(self, output="print") -> Optional[Dict[str, dict]]:
-        """ Summarize changes that were made with respect to the table
+    def summarize_changes(self, output="print") -> dict[str, dict] | None:
+        """Summarize changes that were made with respect to the table
         as loaded from the database.
         If notes/cards/etc. were not loaded at all (and hence also definitely
         not modified), they do not appear in the output.
@@ -163,23 +153,25 @@ class Collection(object):
             as_dict = {}
             for key, value in self.__items.items():
                 if value is not None:
-                    as_dict[key] = value.summarize_changes(output="dict")
-            return as_dict
+                    changes = value.summarize_changes(output="dict")
+                    as_dict[key] = changes  # type: ignore
+            return as_dict  # type: ignore
         elif output == "print":
             for key, value in self.__items.items():
                 if value is not None:
-                    print("======== {} ========".format(key))
+                    print(f"======== {key} ========")
                     value.summarize_changes()
+            return None  # explicit for mypy
         else:
-            raise ValueError("Invalid output setting: {}".format(output))
+            raise ValueError(f"Invalid output setting: {output}")
 
     def _prepare_write_data(
         self, modify=False, add=False, delete=False
-    ) -> Dict[str, Any]:
+    ) -> dict[str, Any]:
         prepared = {}
         for key, value in self.__items.items():
             if value is None:
-                log.debug("Write: Skipping {}, because it's None.".format(key))
+                log.debug("Write: Skipping %s, because it's None.", key)
                 continue
             if key in ["notes", "cards", "revs"]:
                 ndeleted = len(value.was_deleted())
@@ -204,8 +196,9 @@ class Collection(object):
 
                 if not ndeleted and not nmodified and not nadded:
                     log.debug(
-                        "Skipping table {key} for writing, because nothing "
-                        "seemed to have changed".format(key=key)
+                        "Skipping table %s for writing, because nothing "
+                        "seemed to have changed",
+                        key,
                     )
                     continue
 
@@ -214,45 +207,42 @@ class Collection(object):
                     mode = "update"
                 if add and not modify and not delete:
                     mode = "append"
-                log.debug(
-                    "Will update table {key} with mode {mode}".format(
-                        key=key, mode=mode
-                    )
-                )
+                log.debug("Will update table %s with mode %s", key, mode)
                 value.check_table_integrity()
                 raw_table = value.raw()
                 prepared[key] = {"raw": raw_table, "mode": mode}
 
         return prepared
 
-    def _get_and_update_info(self) -> Dict[str, Any]:
-        info = raw.get_info(self.db)
+    def _get_and_update_info(self) -> dict[str, Any]:
+        with closing(self.db) as db:
+            info = raw.get_info(db)
 
-        info_updates = dict(
-            mod=int(time.time() * 1000),  # Modification time stamp
-            usn=-1,  # Signals update needed
-        )
-        if raw.get_db_version(self.db) == 0:
-            for key in info_updates:
-                assert key in info.keys()
-            info.update(info_updates)
-        elif raw.get_db_version(self.db) == 1:
-            assert len(list(info.keys())) == 1
-            first_key = list(info.keys())[0]
-            info[first_key].update(info_updates)
-        # fixme: this currently doesn't work. In the new db structure there's
-        #   a tags table instead of a field, but it doesn't seem to be
-        #   used.
-        # if self.__items["notes"] is not None:
-        #
-        #     missing_tags = list(
-        #         set(info["tags"].keys())
-        #         - set(self.__items["notes"].list_tags())
-        #     )
-        #     for tag in missing_tags:
-        #         # I'm assuming that this is the usn (update sequence number)
-        #         # of the tags
-        #         info["tags"][tag] = -1
+            info_updates = dict(
+                mod=int(time.time() * 1000),  # Modification time stamp
+                usn=-1,  # Signals update needed
+            )
+            if raw.get_db_version(db) == 0:
+                for key in info_updates:
+                    assert key in info
+                info.update(info_updates)
+            elif raw.get_db_version(db) == 1:
+                assert len(info) == 1
+                first_key = list(info)[0]
+                info[first_key].update(info_updates)
+            # fixme: this currently doesn't work. In the new db structure there's
+            #   a tags table instead of a field, but it doesn't seem to be
+            #   used.
+            # if self.__items["notes"] is not None:
+            #
+            #     missing_tags = list(
+            #         set(info["tags"].keys())
+            #         - set(self.__items["notes"].list_tags())
+            #     )
+            #     for tag in missing_tags:
+            #         # I'm assuming that this is the usn (update sequence number)
+            #         # of the tags
+            #         info["tags"][tag] = -1
         return info
 
     def write(
@@ -260,9 +250,9 @@ class Collection(object):
         modify=False,
         add=False,
         delete=False,
-        backup_folder: Union[PurePath, str] = None,
+        backup_folder: PurePath | str | None = None,
     ):
-        """ Creates a backup of the database and then writes back the new
+        """Creates a backup of the database and then writes back the new
         data.
 
         .. danger::
@@ -302,7 +292,7 @@ class Collection(object):
                 modify=modify, add=add, delete=delete
             )
             log.debug("Now getting & updating info.")
-            info = self._get_and_update_info()
+            self._get_and_update_info()
         except Exception as e:
             log.critical(
                 "Something went wrong preparing the data for writing. "
@@ -322,7 +312,7 @@ class Collection(object):
         backup_path = ankipandas.paths.backup_db(
             self.path, backup_folder=backup_folder
         )
-        log.info("Backup created at {}.".format(backup_path.resolve()))
+        log.info("Backup created at %s", backup_path.resolve())
         log.warning(
             "Currently AnkiPandas might not be able to tell Anki to"
             " sync its database. "
@@ -340,25 +330,48 @@ class Collection(object):
         log.debug("Now actually writing to database.")
         try:
             for table, values in prepared.items():
-                log.debug("Now setting table {}.".format(table))
-                raw.set_table(
-                    self.db, values["raw"], table=table, mode=values["mode"]
-                )
-                log.debug("Setting table {} successful.".format(table))
+                log.debug("Now setting table %s.", table)
+                with closing(self.db) as db:
+                    raw.set_table(
+                        db, values["raw"], table=table, mode=values["mode"]
+                    )
+                    log.debug("Setting table %s successful.", table)
+
+                    if table == "cards":
+                        raw.update_card_indices(db)
+
+                    elif table == "notes":
+                        raw.update_note_indices(db)
+
             # log.debug("Now setting info")
             # raw.set_info(self.db, info)
             # log.debug("Setting info successful.")
         except Exception as e:
             log.critical(
-                "Error while writing data to database at {path}"
+                "Error while writing data to database at %s. "
                 "This means that your database might have become corrupted. "
                 "It's STRONGLY advised that you manually restore the database "
-                "by replacing it with the backup from {backup_path} and restart"
+                "by replacing it with the backup from %s and restart"
                 " from scratch. "
                 "Please also open a bug report at "
                 "https://github.com/klieret/AnkiPandas/issues/, as errors "
-                "during the actual writing process should never occur!".format(
-                    path=self.path.resolve(), backup_path=backup_path.resolve()
-                )
+                "during the actual writing process should never occur!",
+                self.path.resolve(),
+                backup_path.resolve(),
             )
             raise e
+        # This is important, because we have written to the database but still
+        # have it opened, which can cause DatabaseErrors.
+        log.debug("I will now drop all copies of the tables")
+        for key in self.__items:
+            self.__items[key] = None
+        for key in self.__original_items:
+            self.__original_items[key] = None
+        log.debug("I will now reload the connection.")
+        self._db = raw.load_db(self.path)
+        log.info(
+            "In case you're running this from a Jupyter notebook, make "
+            "sure to shutdown the kernel or delete all ankipandas objects"
+            " before you open anki to take a look at the result (only one "
+            "application can use the database at a time)."
+        )
