@@ -1,6 +1,6 @@
-#!/usr/bin/env python3
-
 # std
+from __future__ import annotations
+
 import pathlib
 import shutil
 
@@ -13,7 +13,7 @@ from ankipandas.test.util import parameterized_paths
 
 
 def _init_all_tables(col: Collection) -> None:
-    """ Access all attributes at least once to ensure that they are
+    """Access all attributes at least once to ensure that they are
     initialized.
     """
     _ = col.notes
@@ -42,7 +42,7 @@ def test_summarize_changes_no_changes(db_path):
         assert sc[item]["n_modified"] == 0
         assert sc[item]["n_added"] == 0
         assert sc[item]["n_deleted"] == 0
-        assert sc[item]["has_changed"] == False
+        assert not sc[item]["has_changed"]
 
 
 @parameterized_paths()
@@ -62,6 +62,7 @@ def test_read_write_identical_trivial(db_path, tmpdir):
     db_path = shutil.copy2(str(db_path), str(tmpdir))
     (pathlib.Path(str(tmpdir)) / "backups").mkdir()
     col = Collection(db_path)
+    _init_all_tables(col)
     col.write(modify=True, delete=True, add=True)
     col_rel = Collection(db_path)
     assert col.notes.equals(col_rel.notes)
@@ -74,6 +75,7 @@ def test_write_raises_delete(db_path, tmpdir):
     db_path = shutil.copy2(str(db_path), str(tmpdir))
     (pathlib.Path(str(tmpdir)) / "backups").mkdir()
     col = Collection(db_path)
+    _init_all_tables(col)
     col.notes.drop(col.notes.index, inplace=True)
     cases = [
         dict(modify=False, add=True),
@@ -90,6 +92,7 @@ def test_write_raises_modified(db_path, tmpdir):
     db_path = shutil.copy2(str(db_path), str(tmpdir))
     (pathlib.Path(str(tmpdir)) / "backups").mkdir()
     col = Collection(db_path)
+    _init_all_tables(col)
     col.notes.add_tag("test", inplace=True)
     cases = [
         dict(add=False, delete=True),
@@ -106,6 +109,7 @@ def test_write_raises_added(db_path, tmpdir):
     db_path = shutil.copy2(str(db_path), str(tmpdir))
     (pathlib.Path(str(tmpdir)) / "backups").mkdir()
     col = Collection(db_path)
+    _init_all_tables(col)
     col.notes.add_note("Basic", ["test", "back"], inplace=True)
     cases = [
         dict(modify=False, delete=True),
@@ -115,3 +119,13 @@ def test_write_raises_added(db_path, tmpdir):
     for case in cases:
         with pytest.raises(ValueError, match=".*would be modified.*"):
             col.write(**case, add=False)
+
+
+@parameterized_paths()
+def test_write_added(db_path, tmpdir):
+    db_path = shutil.copy2(str(db_path), str(tmpdir))
+    (pathlib.Path(str(tmpdir)) / "backups").mkdir()
+    col = Collection(db_path)
+    _init_all_tables(col)
+    col.notes.add_note("Basic", ["test", "back"], inplace=True)
+    col.write(add=True)
