@@ -1,23 +1,22 @@
-#!/usr/bin/env python3
-
 # std
+from __future__ import annotations
+
 import collections
-import unittest
 import tempfile
-from typing import List
+import unittest
 from pathlib import Path
 
 # 3rd
-from randomfiletree import sample_random_elements, iterative_gaussian_tree
+from randomfiletree import iterative_gaussian_tree, sample_random_elements
 
 # ours
 import ankipandas.paths as paths
-from ankipandas.util.misc import flatten_list_list
 from ankipandas.util.log import set_debug_log_level
+from ankipandas.util.misc import flatten_list_list
 
 
-def touch_file_in_random_folders(basedir, filename: str, n=1) -> List[Path]:
-    """ Create files in random folders.
+def touch_file_in_random_folders(basedir, filename: str, n=1) -> list[Path]:
+    """Create files in random folders.
 
     Args:
         basedir: Starting directory
@@ -98,7 +97,7 @@ class TestFindDatabase(unittest.TestCase):
                     ),
                 )
             )
-            b = sorted(map(str, self.dbs[d]))
+            b = sorted(str(x) for x in self.dbs[d])
             self.assertListEqual(a, b)
 
     def test__find_database_filename(self):
