@@ -1,5 +1,3 @@
-#!/usr/bin/env python3
-
 """ Various utilities of this package.
 
 .. warning::
@@ -9,6 +7,8 @@
 
 """
 
+from __future__ import annotations
+
 import ankipandas.util.checksum
 import ankipandas.util.dataframe
 import ankipandas.util.log
