@@ -1,12 +1,12 @@
-#!/usr/bin/env python3
-
 # std
-from typing import List, Any
+from __future__ import annotations
+
 import collections
+from typing import Any
 
 
 def invert_dict(dct: dict) -> dict:
-    """ Invert dictionary, i.e. reverse keys and values.
+    """Invert dictionary, i.e. reverse keys and values.
 
     Args:
         dct: Dictionary
@@ -19,13 +19,13 @@ def invert_dict(dct: dict) -> dict:
     """
     if not len(set(dct.values())) == len(dct.values()):
         print(dct)
-        print(sorted(list(dct.values())))
+        print(sorted(dct.values()))
         raise ValueError("Dictionary does not seem to be invertible.")
     return {value: key for key, value in dct.items()}
 
 
-def flatten_list_list(lst: List[List[Any]]) -> List[Any]:
-    """ Takes a list of lists and returns a list of all elements.
+def flatten_list_list(lst: list[list[Any]]) -> list[Any]:
+    """Takes a list of lists and returns a list of all elements.
 
     Args:
         lst: List of Lists
@@ -37,7 +37,7 @@ def flatten_list_list(lst: List[List[Any]]) -> List[Any]:
 
 
 def nested_dict():
-    """ This is very clever and stolen from
+    """This is very clever and stolen from
     https://stackoverflow.com/questions/16724788/
     Use it to initialize a dictionary-like object which automatically adds
     levels.
@@ -51,7 +51,7 @@ def nested_dict():
     return collections.defaultdict(nested_dict)
 
 
-def defaultdict2dict(defdict: collections.defaultdict):
+def defaultdict2dict(defdict: collections.defaultdict) -> dict:
     return {
         key: defaultdict2dict(value)
         if isinstance(value, collections.defaultdict)
