@@ -1,14 +1,14 @@
-#!/usr/bin/env python3
-
 # std
+from __future__ import annotations
+
 import unittest
 
 # ours
-from ankipandas.util.log import log, get_logger, set_log_level
+from ankipandas.util.log import get_logger, log, set_log_level
 
 
 class TestLogging(unittest.TestCase):
-    """ Only tests that things run without error. """
+    """Only tests that things run without error."""
 
     def test_log(self):
         log.info("Test info")
